;;; AspectScheme*
;;;  by Éric Tanter
;;; Based on: 
;;; AspectScheme
;;; Copyright (c) 2005, 2006, 2007, 2008 by Christopher Dutchyn (cdutchyn@cs.usask.ca); all rights reserved.
;;;
;;; Extension of AspectScheme using execution levels.
;;; (see "Execution Levels for Aspect-Oriented Programming")
;;; main differences:
;;; - execution of pointcuts and advices happens at a higher level
;;; - no pc/adv join points (not needed anymore, thanks to levels)
;;; - app/prim is not exposed anymore (not needed, thanks to levels)
;;; - exports level-shifting operators ^up and ^down
;;; - exports lambda-s to define level-capturing functions
;;;
;;; + slight difference from paper: an aspect has a "level predicate", rather than just one level
;;; (more flexible, can experiment with >= semantics)
;;;
;;; most changes to AspectScheme are identifiable by the ";;LEVEL" or ";;ET" comment notes.

#lang scheme

(require (only-in swindle/base let let* letrec))

(provide (except-out (all-from-out scheme) lambda #%app)
         (rename-out [app/weave #%app] ;;LEVEL: no app/prim
                     [lambda/aspects lambda]
                     [lambda/aspects/capture lambda-s]) ;;LEVEL
         fluid-around
         around 
         toplevel-around
         && || !
         top? top below above bottom bottom?
         target args some-args
         call? exec? call exec ;;LEVEL: no pc/adv
         cflowtop cflowbelow cflowbottom cflowabove
         cflow within
         let let* letrec
         ^up ^down ;;LEVEL
         )


;; Join Point
;;                Proc           args
;; jp ::= call-jp a->b             a	;; procedure application ('a' can be values (ie. tuple {...})
;;     |  exec-jp a->b             a	;; procedure execution (cannot be advised only matched)
;;   AspectJ matches and transforms dispatches and calls, but not executions
;;   they're just poorly named (dispatch == `call', call == `execution')
;;     |  adv-jp  (a->b)->c->a->b  c	;; advice execution ... 'c' can be values as well

;; Pointcut
;; pc :: {[jp]*jp*[jp]}->c		;; above * jp * below

;; Advice
;; adv :: (a->b)->c->a->b

;; Aspect
;; aspect ::=    fluid-around pc adv body  ;; dynamic scoping
;;         |           around pc adv body  ;; lexical scoping
;;         |  toplevel-around pc adv       ;; top-level scoping (i.e. body is rest of repl)

;; aspect structure
(define-struct aspect (pc adv levelp)) ;; LEVEL

(define-struct          jp (target args level)) ;;LEVEL
(define-struct (call-jp jp)(           ))
(define-struct (exec-jp jp)(           ))


;;; current execution level is a parameter
(define current-level (make-parameter 0)) ;; LEVEL

;;; level-shifting operators ;; LEVEL
;; up
(define-syntax ^up
  (lambda (stx)
    (syntax-case stx ()
      [(_ body ...)
       (syntax/loc stx (parameterize ((current-level (+ 1 (current-level))))
                         body ...))])))

;; down
(define-syntax ^down
  (lambda (stx)
    (syntax-case stx ()
      [(_ body ...)
       (syntax/loc stx (parameterize ((current-level (- (current-level) 1)))
                         body ...))])))

;; PLT Scheme modules fail with fluid-let and regular definitions so we must use parameters
;;   the error is "set!: cannot mutate module-required variable",
;;  c.f. http://list.cs.brown.edu/pipermail/plt-scheme/2004-September/006723.html
;;  n.b. using dynamic-wind to provide fluid-let is subtle
(define-syntax fluid-let-parameter
  (syntax-rules ()
    [(_ ([p v]) e ...)
     (let ([y v])
       (let ([(swap) (let ([t (p)])
                       (p y)
                       (set! y t))])
         (dynamic-wind swap
                       (lambda () e ...)
                       swap)))]))

;; dynamically-scoped aspects
;; NB. We use dynamic binding to illuminate the connection to the
;; simplified semantics in the Science of Computer Programming paper
;; where dynamic-scoped variables hold static and dynamic aspects.
;; As noted above, fluid-let fails, so we use the next-best thing.
(define dynamic-aspects (make-parameter '()))
(define  static-aspects (make-parameter '()))

(define-syntaxes (fluid-around around)
  (let ([round (lambda (param)
                 (lambda (stx)
                   (syntax-case stx ()
                     [(_ pc adv levelp body ...)
                      (quasisyntax/loc stx 
                        (fluid-let-parameter ([#,param (cons (make-aspect pc adv levelp) (#,param))])
                          body ...))])))])
    (values (round #`dynamic-aspects)   ;dynamically-scoped
            (round #`static-aspects)))) ;lexically-scoped

;; lexically-scoped aspects
(define-syntax lambda/aspects ;;ET: renamed  
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx (let ([aspects (static-aspects)])
                         (lambda params
                           (fluid-let-parameter ([static-aspects aspects])
                             body ...))))])))

;; defining level-capturing functions
(define-syntax lambda/aspects/capture
  (lambda (stx)
    (syntax-case stx ()
      [(_ params body ...)
       (syntax/loc stx (let ([aspects (static-aspects)]
                             [level (current-level)]) ; get current level
                         (lambda params
                           (fluid-let-parameter ([static-aspects aspects])
                             (fluid-let-parameter ([current-level  level]) ; reinstate for body eval
                               body ...)))))])))

;; top-level aspects
(define toplevel-aspects (make-parameter '()))

(define (toplevel-around pc adv levelp)
  (toplevel-aspects (cons (make-aspect pc adv levelp) (toplevel-aspects))))

;; weaver
;; current aspects -- in decending order of application!
(define (current-aspects)
  (append (dynamic-aspects)
          (static-aspects)
          (toplevel-aspects)))

;; join points implemented as continuation marks
;;  (AKA fluid-let that respects safe-for-space properties)

;; ensure that our continuation mark is unique and unforgeable
(define-struct jp-mark (id))
(define jp-mark-tag (make-jp-mark 'joinpoint))
  
(define (jp-context) 
  (continuation-mark-set->list
   (current-continuation-marks) 
   jp-mark-tag))

(define-syntax with-joinpoint
  (syntax-rules ()
    [(_ jp body ...)
     ((lambda (x) x)
      (with-continuation-mark jp-mark-tag jp
        (begin body ...)))]))

;; replacement for #%app
(define-syntax app/weave
  (lambda (stx)
    (syntax-case stx ()
      [(_ f a ...)
       (syntax/loc stx (app/weave/rt f a ...))])))

(define (app/weave/rt fun-val . arg-vals)
  (if (primitive? fun-val)
      (apply fun-val arg-vals)
      (let ([jp (make-call-jp fun-val arg-vals (current-level))] ;;LEVEL
            [jp+ (jp-context)])
        (with-joinpoint jp              
          (let ((base-eval (lambda arg-vals
                              (with-joinpoint (make-exec-jp fun-val arg-vals (current-level)) ;;LEVEL
                                              (^down (apply fun-val arg-vals))))));; proceed at -1

            (^up 
             (apply  (weave base-eval '() jp jp+ (current-aspects)) 
                     arg-vals)))))))

(define (weave fun-val jp- jp jp+ aspects)
  (foldr (lambda (aspect fun)
           (cond
             [(and ((aspect-levelp aspect) (jp-level jp))  
                   (with-joinpoint (make-call-jp (aspect-pc aspect) (list jp- jp jp+) (current-level)) ;;LEVEL
                                   ((aspect-pc aspect) jp- jp jp+)))                                   ; plain call jp
              => (lambda (ctxt-vals)
                   (lambda args
                   (with-joinpoint (make-call-jp (aspect-adv aspect) ctxt-vals (current-level)) ;;LEVEL
                                   (apply (apply ((aspect-adv aspect) fun) ctxt-vals) args))))] ; plain call jp
             [else fun]))                                                                       ;;ET: fixed bug in jp gen
         fun-val
         aspects))

;; pointcuts -- strict combinators
;; NB. This PLT Scheme module does not export app/weave for #%app
;; until the end of the module, so these definitions do not
;; require app/prim.
(define ((&& . pcs) jp- jp jp+)
  (let loop ([pcs  pcs]
             [res  '()])
    (if (null? pcs)
        res
        (let ([r ((car pcs) jp- jp jp+)])
          (and r
               (loop (cdr pcs) (append res r)))))))

(define ((|| . pcs) jp- jp jp+)
  (let loop ([pcs pcs])
    (and (not (null? pcs))
         (or ((car pcs) jp- jp jp+)
             (loop (cdr pcs))))))

(define ((! pc) jp- jp jp+)
  (and (not (pc jp- jp jp+))
       '()))

;; pointcuts -- structural
(define (top? jp- jp jp+)
  (and (null? jp+)
       '()))

(define ((below pc) jp- jp jp+)
  (and (not (null? jp+))
       (pc (cons jp jp-) (car jp+) (cdr jp+))))

(define ((above pc) jp- jp jp+)
  (and (not (null? jp-))
       (pc (cdr jp-) (car jp-) (cons jp jp+))))

(define (bottom? jp- jp jp+)
  (and (null? jp-)
       '()))

;; pointcuts -- `binding'
(define (target jp- jp jp+)
  (list (jp-target jp)))

(define (args jp- jp jp+)
  (jp-args jp))

(define ((some-args as) jp- jp jp+)
  (foldl (lambda (a v l)
           (if a
               (cons v l)
               l))
         '()
         as
         (jp-args jp)))

;; pointcuts -- fundamental
(define ((true-pc) jp- jp jp+)
  '())

(define ((this-jp? p?) jp- jp jp+)
  (and (p? jp)
       '()))

(define-values (call? exec?)     ;; LEVEL: removed pc/adv
  (values (this-jp? call-jp?)
          (this-jp? exec-jp?)))

(define-values (call exec)       ;; LEVEL: removed pc/adv
  (let* ([(target= f) (this-jp? (lambda (jp)
                                      (eqv? f (jp-target jp))))]
         [((this-target? p) f) (&& (target= f)
                                  p)])
    (values (this-target? call?)
            (this-target? exec?))))

(define (display-jps jp- jp jp+)
  (for-each (lambda (jp) (display `(< ,jp ,(jp-args jp)))) jp-)
  (display `(= ,jp ,(jp-args jp)))
  (for-each (lambda (jp) (display `(> ,jp ,(jp-args jp)))) jp+)
  (newline))

;; pointcuts - higher-order recursive
(define (((cflow-walk pc) step end) jp- jp jp+)
  ((|| pc
       (&& (! end)
           (step ((cflow-walk pc) step end)))) jp- jp jp+))

(define (top pc)
  (&& pc
      (! (cflowbelow pc))))

(define (bottom pc)
  (&& pc
      (! (cflowabove pc))))

;; pointcuts - higher-order points-free
(define (cflowtop pc)
  (cflowbelow (top pc)))

(define (cflowbelow pc)
  (below ((cflow-walk pc) below top?)))

(define (cflowabove pc)
  (above ((cflow-walk pc) above bottom)))

(define (cflowbottom pc)
  (cflowbelow (bottom pc)))

;; pointcuts - compatibility
(define (cflow pc)
  ((cflow-walk pc) below top?))

;; this one is interesting -- can you do this without cflowabove?
;; I think not; it is a special case of enclosingexecution where
;; no other executions are allowed in between _here_ and the execution
;; ie. we have a dynamic test for a lexical condition
;; of course, tail call optimization makes executions disappear, so
;; we presume that TCO for `interesting' executions does happen
;; Furthermore, since `interesting' (== tested for in a pointcut)
;; executions might not be known until after we've accumulated a context
;; with them
;;  (let ([f (lambda (g)
;;             (let ([h (lambda () 1)])
;;               (around (&& (call h)
;;                           (cflow (call g)))
;;                       (lambda (jp)
;;                         (lambda ()
;;                           (+ 1 (jp))))
;;                     (h))))])
;;    (f f))
;; then the problem becomes insurmountable in the general case.
;; So what specific cases can be optimized (modulo tail calls)?

(define (within f)
  (cflowbelow (&& (exec f)
                  (! (cflowabove call?)))))
